
/***************************************************************************
 * IBM Confidential
 *--------------------------------------------------------------------------
 * IBM Haifa Research Lab.
 *--------------------------------------------------------------------------
 *
 * Product      : ReGS Reference Implementation
 * Filename     : $RCSfile: Reader.cpp,v $
 *
 * Description  : EvenLogReader receives target event log name and output file as 
 *				  parameters and reads events from Windows using Windows EventLog API. 
 *				  The events are then transformed into strings and written to the 
 *				  standard output and to the file.
 *
 * @author      : Arie Orlovsky (ARIEO@il.ibm.com)
 * @version     : $Revision: 1.2 $
 * Modified on  : $Date: 2005/02/04 07:20:01 $ by $Author: schiluku $
 * Modified on  : $Date: 2005/02/04 07:20:01 $ by $Author: schiluku $
 * Modified on  : $Date: 2005/02/04 07:20:01 $ by $Author: schiluku $
					Modified to get full lenth description string as displayed in 
					Windows Event Viewer.
 * Modified on  : $Date: 2005/02/04 07:20:01 $ by $Author: schiluku $
					Changed function names and formatting for readability. Added clean up 
					code to handle memory leaks.
 *--------------------------------------------------------------------------
 * DISCLAIMER
 *  - This code is provided AS IS.
 *  - The code is still experimental and may exhibit undesired behavior
 ***************************************************************************/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <winerror.h>

const LPSTR cRecordNumberPref = "RecordNumber: ";
const LPSTR cEventTypePref = "Severity: ";
const LPSTR cCreationTimePref = "CreationTime: ";
const LPSTR cSourceComponentPref = "SourceComponent: ";
const LPSTR cEventCategoryPref = "EventCategory: ";
const LPSTR cEventIDPref = "EventID: ";
const LPSTR cUsernamePref = "Username: ";
const LPSTR cComputerNamePref = "Computername: ";
const LPSTR cLocationTypePref = "LocationType: ";
const LPSTR cLocationType = "Hostname";
const LPSTR cMessagesPref = "Message: ";

const LPSTR cErrorHeader = "<ACADErrorHeader>";

const int cbeDEBUG = 10;
const int cbeINFO = 20;
const int cbeWARN = 30;
const int cbeERROR = 40;
const int cbeFATAL = 50;

const int MAX_MSG_LENGTH = 1024;

FILE *fpOut = NULL;
// FILE* hFile;

LPSTR GetDescriptionStringFromRegKey(EVENTLOGRECORD *pRecord,  LPSTR source);
LPSTR getSourceName(EVENTLOGRECORD *record);
BOOL ReadEventSourceInfo(LPCSTR lpszESName, LPSTR lpszEvent);


/* eventmessagefile handle 
   event description message index
   Language ID of the message
   Array of insertion strings */
LPSTR GetEventMessage(HMODULE hDll, DWORD dwEventIndex, DWORD dwLanguageID, LPTSTR *lpInserts );


int getEventType(WORD ntEventType)
{
	int result = 0;
	switch (ntEventType)
	{
	case (EVENTLOG_ERROR_TYPE):
		result = cbeERROR;
		break;
	case (EVENTLOG_WARNING_TYPE):
		result = cbeWARN;
		break;
	case (EVENTLOG_INFORMATION_TYPE):
		result = cbeINFO;
		break;
	case (EVENTLOG_AUDIT_SUCCESS):
		result = 15;
		break;
	case (EVENTLOG_AUDIT_FAILURE):
		result = 16;
		break;
	default :
		result = 17;
		break;
    }

    return result;
}

void error(LPSTR errMessage)
{
	printf("%s\n%s\n", cErrorHeader, errMessage);
}


// removing carriage returns '\r' as well as line feeds '\n'
void removeNewLine(LPSTR str)
{
	for (int i=0; str[i] != '\0'; i++)
	{
		if (str[i] == '\n' || str[i] == '\r')
		{
			//str[i] = '&#0x0A;';
			str[i] = ' ';
		}
	}
}

void GetDescriptionString(LPSTR *pMessage, EVENTLOGRECORD *record)
{
	*pMessage = GetDescriptionStringFromRegKey(record, getSourceName(record));
	// the message retrieved contains the description text displayed in the Event Viewer
	// we do not need to further get any messages.
	if(*pMessage != NULL && strlen(*pMessage) > 2)
	{
		// remove if the string contains any new lines otherwise the adapter will fail.
		removeNewLine(*pMessage);
		return;
	}

	// Since the description text could not be retrieved get the message strings
	*pMessage = new char[(record->DataOffset - record->StringOffset) + 5];
	DWORD recordStringStart = (DWORD)record + record->StringOffset;
	LPSTR recordString = (LPSTR)recordStringStart;
	// initialize the array to nulls
	memset(*pMessage, NULL, (record->DataOffset - record->StringOffset));

	for (int i=0; i < record->NumStrings; i++)
	{
		removeNewLine(recordString);
		strcat(*pMessage, recordString);
		strcat(*pMessage, " ");
		recordStringStart = recordStringStart + strlen(recordString) + 1;
		recordString = (LPSTR)recordStringStart;
	}
}

void copyTimeGenerated(LPSTR timeBuffer, PEVENTLOGRECORD pevlr)
{
	FILETIME FileTime, LocalFileTime;
	SYSTEMTIME SysTime;
	__int64 lgTemp;
	__int64 SecsTo1970 = 116444736000000000;

	lgTemp = Int32x32To64(pevlr->TimeGenerated,10000000) + SecsTo1970;

	FileTime.dwLowDateTime = (DWORD) lgTemp;
	FileTime.dwHighDateTime = (DWORD)(lgTemp >> 32);

	FileTimeToLocalFileTime(&FileTime, &LocalFileTime);
	FileTimeToSystemTime(&LocalFileTime, &SysTime);

	sprintf(timeBuffer,"%02d/%02d/%02d %02d:%02d:%02d.%06d",
		SysTime.wMonth,
		SysTime.wDay,
		SysTime.wYear,
		SysTime.wHour,
		SysTime.wMinute,
		SysTime.wSecond,
		SysTime.wMilliseconds);
}

LPSTR getSourceName(EVENTLOGRECORD *record)
{
	return (LPSTR) ((LPSTR)record + sizeof(EVENTLOGRECORD));
}

LPSTR getComputerName(EVENTLOGRECORD *record)
{
    LPSTR sourceName = getSourceName(record);
    return (LPSTR) (sourceName + strlen(sourceName) + 1);
}

void cpyUserName(LPSTR sidBuffer, DWORD sidLength, EVENTLOGRECORD *record)
{
	if (record->UserSidLength == 0)
	{
		strcpy(sidBuffer, "N/A");
		return;
	}

	char referencedDomainName[500];
	DWORD domainLength = sizeof(referencedDomainName);
	SID_NAME_USE eUse;
	int res;

	res = LookupAccountSid( NULL, // [input] system name
			    (SID *)((LPSTR)record + record->UserSidOffset), // [input] the SID structure
			    sidBuffer, // [output] the buffer receiving the account name
			    &sidLength, // [input/output] the buffer length for the account name
			    referencedDomainName, // [output] domain name
			    &domainLength, // [input/output] the buffer length for the domain name
			    &eUse // [output] the type of the account
			    );
	if (res == 0)
	{
		strcpy(sidBuffer, "N/A");
		return;
	}
}

// writting EventLog event to the standard ouput
BOOL writeNextRecord(EVENTLOGRECORD *pevlr)
{
	LPSTR sourceName = getSourceName(pevlr);
	LPSTR compName = getComputerName(pevlr);

	const DWORD cUsernameSize = 500;
	char username[cUsernameSize];
	DWORD userNameSize = cUsernameSize;

	cpyUserName(username, userNameSize, pevlr);

	LPSTR stringsBuffer = NULL;
	GetDescriptionString(&stringsBuffer, pevlr);

	const DWORD cTimeSize = 500;
	char time[cTimeSize];
	DWORD timeSize = cTimeSize;

	copyTimeGenerated(time, pevlr);

	int CBEseverity = getEventType(pevlr->EventType);

	DWORD dwEventId = pevlr->EventID;
	DWORD dwMSB16Zeros = 65535; // this double word variable has its first 16 bits as zero and the remaining as 1
	dwEventId &= dwMSB16Zeros;

	if(printf("%s%d;\t%s%d;\t%s%s;\t%s%s;\t%s%d;\t%s%08X;\t%s%s;\t%s%s;\t%s%s;\t%s%s;\n" ,cRecordNumberPref,
		pevlr->RecordNumber, cEventTypePref, CBEseverity, cCreationTimePref,time,
		cSourceComponentPref, (LPSTR) ((LPBYTE) pevlr + sizeof(EVENTLOGRECORD)),
		cEventCategoryPref, pevlr->EventCategory, cEventIDPref, dwEventId,
		cUsernamePref, username, cComputerNamePref, compName, cLocationTypePref, cLocationType,
		cMessagesPref, stringsBuffer) < 0)
	{
		error("Failed to write event to output");
		return FALSE;
	}

	if(fpOut)
	if(fprintf(fpOut, "%s%d;\t%s%d;\t%s%s;\t%s%s;\t%s%d;\t%s%08X;\t%s%s;\t%s%s;\t%s%s;\t%s%s;\n" ,cRecordNumberPref,
		pevlr->RecordNumber, cEventTypePref, CBEseverity, cCreationTimePref,time,
		cSourceComponentPref, (LPSTR) ((LPBYTE) pevlr + sizeof(EVENTLOGRECORD)),
		cEventCategoryPref, pevlr->EventCategory, cEventIDPref, dwEventId,
		cUsernamePref, username, cComputerNamePref, compName, cLocationTypePref, cLocationType,
		cMessagesPref, stringsBuffer) < 0)
	{
		error("Failed to write event to output file");
		return FALSE;
	}

	_flushall(  );
	return TRUE;
}

void main(int argc , char * argv[] )
{
	const DWORD cBufferSize = 500;
	HANDLE hLog;
	EVENTLOGRECORD *pevlr = NULL;
	//BYTE bBuffer[cBufferSize];
	DWORD dwRead, dwNeeded;

	//pevlr = (EVENTLOGRECORD *) &bBuffer;

	if(argc < 2)
	{
		error("Usage: EventLogReader.exe logName <outputFilename>");
		return;
	}

	fpOut = fopen(argv[2], "w");
	if(!fpOut)
	{
		//error("Failed to open output file");
		//return;
	}
	//hFile = fopen("d:\\Leva.txt", "w");  // This was for debugging only
	// creating event which will be signaled by the the EventLog
	HANDLE hEvent = CreateEvent(
							NULL,
							FALSE,                      // reset type
							FALSE,                      // initial state
							NULL                        // object name
						);

	if(hEvent == NULL)
	{
		error("Failed to create event");
		return;
	}

	 // Open the Application event log.
	hLog = OpenEventLog( NULL,    // use local computer
             argv[1]);   // source name

	if (hLog == NULL)
	{
		error("Failed to open event log");
		return;
	}

	// subscribing to the events of new entries to the event log
	if(!NotifyChangeEventLog(hLog, hEvent))
	{
		error("FAILED to subscribe for notifications");
		return;
	}

	char one[1];
	LPSTR buffer = NULL;

	//while(1)
	{

		while (1)
		{
			// dummy read to get the size of 1 EventLog entry.
			if(!ReadEventLog(hLog,
					EVENTLOG_SEQUENTIAL_READ|EVENTLOG_FORWARDS_READ,
					0,
					(LPVOID)one,
					1,
					&dwRead,
					&dwNeeded))

			{
				HRESULT hr = GetLastError();
				if(hr == ERROR_HANDLE_EOF)
				{
					//break;
					if(fpOut)
					fclose(fpOut);
					return;
                }

				if(hr != ERROR_INSUFFICIENT_BUFFER)
				{
					error("Failed to read EventLog entry");
					if(fpOut)
					fclose(fpOut);
					return;
				}
			}

			buffer = new char[dwNeeded];
			DWORD temp = dwNeeded;
			if (buffer == NULL) {
				printf("cannot allocate memory");
				//break;
				if(fpOut)
				fclose(fpOut);
				return;
			}

			//reading 1 entry from the event log
			if(!ReadEventLog(hLog,                // event log handle
					EVENTLOG_SEQUENTIAL_READ|EVENTLOG_FORWARDS_READ, // sequential read
					0,              // ignored for sequential reads
					(LPVOID)buffer, // address of buffer for read data
					temp,          // size of buffer
					&dwRead,      // number of bytes read
					&dwNeeded))   // bytes in next record
			{
				HRESULT hr = GetLastError();
				error("Failed to read EventLog entry");
				if(fpOut)
				fclose(fpOut);
				return;
			}

			pevlr = (EVENTLOGRECORD *) buffer;

			if(!writeNextRecord(pevlr))
			{
				if(fpOut)
			 	fclose(fpOut);
				return;
            }
		}

		// wait for EventLog notification
//		WaitForSingleObject(hEvent, INFINITE);
	}

	CloseEventLog(hLog);
	if(fpOut)
			 fclose(fpOut);
}


//-------------------

LPSTR GetDescriptionStringFromRegKey(EVENTLOGRECORD *pRecord,  LPSTR source)
{
    BOOL bResult;
    TCHAR szEvent[256],szBuffer[256], **first_sz;
    HMODULE hEvt;
    LPTSTR lpP[21];
    LPSTR    lpBuf = NULL;
    LPTSTR  lpstrlpBuf = "";
    LPSTR pStr = NULL;

    if (pRecord->NumStrings ) 
	{
        pStr = (LPSTR)((LPBYTE)pRecord + pRecord->StringOffset);
    } 
	else 
	{
        pStr = "";
    }

    if ( pStr )
    {
        DWORD i;
        for ( i = 0; i < pRecord->NumStrings; i++ )
        {
            lpP[i] = (LPSTR)pStr;
            pStr = strchr( (LPSTR)pStr, '\0' ) + 1; 
        }
    }
    //Get the file name(s) from the registry
    bResult = ReadEventSourceInfo( source, szEvent);
    if(strchr(szEvent, ';')) 
    {
        int i=0, j, k, num_files=0, last=0;
        LPSTR aux, cad;
        aux = szEvent;
        while(aux = strchr(aux, ';'))
        {
            num_files++;
            aux++;
        }
        aux = szEvent;

        for(j=0; j<= num_files + 1; j++)
        {
            int counter =0;
            while((szEvent[i] != ';' && szEvent[i] != '\0') && i )
			{
                i++;
                counter++;
            }

			if(szEvent[i] == ';' || szEvent[i] == '\0')
			{
				counter++;
				i++;
			}


            first_sz= (TCHAR **)malloc(num_files * (sizeof(szEvent)));
            cad = (LPSTR) malloc(counter+1);

			// get the first file name
			for(k = 0; k < counter; k++ )
			{
                cad[k] = szEvent[last+k];
            }
            cad[k]= '\0';

			last = i;
            i++;

			if(k == 0)
			{
				continue;
			}

            first_sz[j] = cad;
            //Replace the %SystemRoot% kind of environment variables with their values
            ExpandEnvironmentStrings(first_sz[j],szBuffer, 257);

            // Load the EventMessageFile, could be an .exe or .dll
            hEvt = LoadLibraryEx(szBuffer, NULL, DONT_RESOLVE_DLL_REFERENCES );

            /* Load the event message file DLL */ 
            if ( hEvt )
            {
                /* Get the event message with the paramater strings inserted */
                lpBuf = GetEventMessage( hEvt, pRecord->EventID, MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT), lpP );
                FreeLibrary( hEvt ); // unload event message file
            }

			if(lpBuf != NULL && strlen(lpBuf) > 0)
			{
				break;
			}
        }
        return (lpBuf);
    }
    else // single EventMesssageFile found
    {
        ExpandEnvironmentStrings(szEvent, szBuffer, 257);
        hEvt = LoadLibraryEx(szBuffer, NULL, DONT_RESOLVE_DLL_REFERENCES );
        /* Load the event message file DLL */ 
        if ( hEvt )
        {
            /* Get the event message with the paramater strings inserted */
            lpBuf = GetEventMessage( hEvt, pRecord->EventID, MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT), lpP );
            FreeLibrary( hEvt );
        }
    }
    return (lpBuf);
}

//
BOOL ReadEventSourceInfo(LPCSTR lpszESName, LPSTR lpszEvent) 
{
    BOOL bResult = FALSE;
    HKEY hKey;
    DWORD dwBytesReturned;
    TCHAR szKeyName[128];

    /* search for event source registry key */
    lstrcpy( szKeyName, "System\\CurrentControlSet\\Services\\EventLog\\Security\\" );
    lstrcat( szKeyName, lpszESName );
    if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, szKeyName, 0, KEY_ALL_ACCESS, &hKey ) != ERROR_SUCCESS )
    {
        lstrcpy( szKeyName, "System\\CurrentControlSet\\Services\\EventLog\\System\\" );
        lstrcat( szKeyName, lpszESName );
        if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, szKeyName, 0, KEY_ALL_ACCESS, &hKey ) != ERROR_SUCCESS )
        {
            lstrcpy( szKeyName, "System\\CurrentControlSet\\Services\\EventLog\\application\\" );
            lstrcat( szKeyName, lpszESName );
            if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, szKeyName, 0, KEY_ALL_ACCESS, &hKey ) != ERROR_SUCCESS )
            {
				/* The registry key of event source is not found, so message cannot be retrieved. 
				   Things like this are generally encountered when the application that wrote the event to 
				   windows events is uninstalled.
				*/
                return FALSE;
            }
        }
    }
    bResult = TRUE; /* Found the registered event source key */

	/* Get the EventMessageFile path */
    dwBytesReturned = MAX_PATH; // max length of the file path in Windows
    if ( RegQueryValueEx( hKey, "EventMessageFile", NULL, NULL, (LPBYTE)lpszEvent, &dwBytesReturned ) != ERROR_SUCCESS )
	{
		lpszEvent[0] = '\0';
	}

    return( bResult );
}

//Format the message
LPSTR GetEventMessage(HMODULE hDll, DWORD dwEventIndex, DWORD dwLanguageID, LPTSTR *lpInserts )      
{
  DWORD  dwReturn;
  LPSTR lpMsgBuf = NULL;
  DWORD  dwFlags = FORMAT_MESSAGE_FROM_HMODULE |FORMAT_MESSAGE_ALLOCATE_BUFFER;

  if ( lpInserts )
  {
	  dwFlags |= FORMAT_MESSAGE_ARGUMENT_ARRAY; 
  }

  dwReturn = FormatMessage(dwFlags, hDll, dwEventIndex, dwLanguageID, (LPTSTR) &lpMsgBuf, 0, lpInserts );

  if(dwReturn == 0) // message could not be retrieved
  {
	  return NULL;
  }

  return( lpMsgBuf );
} 
